import code
import readline
import atexit
import os
from types import MethodType

class PERMConsole(code.InteractiveConsole):
    def __init__(self, locals = None, filename = '<console>',
                histfile = os.path.expanduser("~/.permm-history")):
        code.InteractiveConsole.__init__(self)
        self.init_history(histfile)
    
    def init_history(self, histfile):
        readline.parse_and_bind("tab: complete")
        if hasattr(readline, "read_history_file"):
            try:
                readline.read_history_file(histfile)
                readline.set_history_length(1000)
            except IOError:
                pass
            atexit.register(self.save_history, histfile)
            
    def save_history(self, histfile):
        readline.write_history_file(histfile)

def load_environ(mech, locals_dict):
    if not locals_dict.has_key('mech'):
        locals_dict['mech'] = mech
    locals_dict.update(mech.species_dict)
    locals_dict.update(mech.reaction_dict)
    try:
        locals_dict.update(mech.irr_dict)
        locals_dict.update(mech.nreaction_dict)
    except:
        pass
    try:
        locals_dict.update(mech.process_dict)
    except:
        pass

    locals_dict.update([(k,getattr(mech,k)) for k in dir(mech) if '__' not in k and isinstance(getattr(mech,k),MethodType) and k not in ('set_mrg', 'set_irr', 'set_ipr')])
    



from yaml import load
from glob import glob
from os.path import basename, abspath, dirname, join
from collections import defaultdict

from permm.core.Mechanism import Mechanism
_mechanisms_dir = abspath(dirname(__file__))

atoms = load(file(join(_mechanisms_dir, 'atoms.yaml')))

class _mech_fromkey(defaultdict):
    def __init__(self):
        self._paths = dict([(basename(path)[:-5], path) for path in glob(join(_mechanisms_dir, '*.yaml'))])

    def keys(self):
        return [k for k in self._paths.keys()]
        
    def iterkeys(self):
        for key in self._paths.keys():
            return key
        
    def iteritems(self):
        for key in self._paths.keys():
            yield key, self[key]

    def __iter__(self):            
        return self.iterkeys()
        
    def keys(self):
        return tuple(self._paths.keys())
        
    def __missing__(self, key):
        if key not in self._paths.keys():
            raise KeyError("%s not in mechanisms")
        
        mech = self[key] = Mechanism(self._paths[key])
        return mech
        
mechanism_dict = _mech_fromkey()


import sys
from collections import defaultdict
from warnings import warn
import unittest

from numpy import ndarray, \
                  dtype, \
                  array, \
                  newaxis, \
                  rollaxis, \
                  arange

from PseudoNetCDF.sci_var import PseudoNetCDFVariable

from Species import Species

def Processes_ProcDelimSpcDict(proc_names, variables, delim = '_'):
    return dict([(k, Process_ProcDelimSpcDict(k, variables, delim = delim)) for k in proc_names])

def Process_ProcDelimSpcDict(proc_name, variables, delim = '_'):
    """
    Create a single process (proc_name) for all species in the variables
    dictionary where each species/process combination is variable
    with the key <proc_name><delim><spec name>
    proc_name - string process name 
    variables - dictionary with '%s%s%s' % (name, delim, speciesname)
    delim - delimiter
    """
    prefix = '%s%s' % (proc_name, delim)
    lp = len(prefix)
    proc_keys = [k for k in variables.keys() if k[:lp] == prefix]
    process_variables = dict([(k[lp:], variables[k]) for k in proc_keys])
    return Process(proc_name, **process_variables)

class Process(object):
    def __init__(self, name, units = {}, default_unit = None, **kwds):
        self.name = name
        self.names = (name,)
        self.__units = units.copy()
        self.data = {}
        for k in kwds.keys():
            self.data[k] = kwds[k][:]
            self.__units[k] = (getattr(kwds[k], 'units', units.get(k, default_unit or 'Unknown'))).strip()
            
    def __getitem__(self, key):
        units = self.__units
        if isinstance(key, Species):
            values = array([self.data[spc] * prop['stoic'] for spc, prop in key.spc_dict.items()]).sum(0)
            units = dict([(k, self.__units[k]) for k in key.spc_dict.keys()])
        elif key in self.data.keys():
            values = self.data[key]
        else:
            new_data = dict([(k, v[key]) for k, v in self.data.iteritems()])
            return Process(self.name, self.__units, **new_data)
        return PseudoNetCDFVariable(self, self.name, values.dtype.char, ('TSTEP',), values = values, units = units)

    def __setitem__(self, key, value):
        return self.data.__setitem__(key, value)
    
    def keys(self):
        return self.data.keys()
        
    def set_units(self, k, units):
        if isinstance(k, Species):
            if set([k.name]) == set(k.names()) and not k.exclude:
                self.__units[k.name] = units
            else:
                result = set([self.__units[key] for key in k.names() if self.data.has_key(key)])
                if len(result) == 1:
                    self.__units[k.name] = units
                else:
                    self.__units[k.name] = 'mixed'
        else:
            self.__units[k] = units
        
    def get_units(self, k, defaultunits = None):
        if isinstance(k, Species):
            if self.__units.has_key(k.name):
                return self.__units.get(k.name, defaultunits)
            else:
                result = set([self.__units[key] for key in k.names() if self.data.has_key(key)])
                if len(result) == 1:
                    return result.pop()
                else:
                    return defaultunits or 'mixed'
        else:
            return self.__units[k]
        
    def __getslice__(self, *args, **kwds):
        """
        Use standard ndarray.__getitem__, but retain
        NetRxnArray type
        """
        result = Process(self.name, self.__units, **dict([('%s' % (k), v.__getslice__(*args, **kwds)) for k, v in self.data.iteritems()]))
        return result

    def __generic_math_operator(self, rhs, operation):
        operator = {'+': ndarray.__add__,
                    '-': ndarray.__sub__,
                    '/': ndarray.__div__,
                    '*': ndarray.__mul__
                   }[operation]
        if isinstance(rhs, self.__class__):
            result = Process(name = '%s %s %s' % (self.name, operation, rhs.name))
            result.names = self.names + rhs.names
            for key in self.keys():
                result[key] = self.data[key].copy()
                result.set_units(key, self.get_units(key))
                
            for key in rhs.keys():
                rhs_unit = rhs.get_units(key)
                if result.data.has_key(key):
                    result[key] = operator(result[key], rhs[key])
                    result_unit = result.get_units(key)
                    if rhs_unit != result_unit:
                        result.set_units(key, '%s %s %s' % (result_unit, operation, rhs_unit))
                else:
                    result[key] = rhs[key].copy()
                    result.set_units(key, rhs_unit)
        else:
            result = Process(name = self.name, units = self.__units.copy())
            for key in self.keys():
                try:
                    result[key] = operator(self[key], rhs)
                except:
                    raise TypeError, "It is unclear how to %s IPR and %s; scalars, arrays, species and processes should have meaningful results" % (operation, type(rhs))
            
        return result
        
    def __add__(self,rhs):
        return self.__generic_math_operator(rhs, '+')
        
    def __sub__(self,rhs):
        return self.__generic_math_operator(rhs, '-')
        
    def __div__(self,rhs):
        return self.__generic_math_operator(rhs, '/')
        
    def __mul__(self,rhs):
        return self.__generic_math_operator(rhs, '*')
        
    def sum(self):
        """
        Sum species over time
        """
        result = Process(name = self.name)
        for key in self.keys():
            result[key] = array(self.data[key].sum(), ndmin = 1)
        
        return result
    
class TestProcess(unittest.TestCase):
    def setUp(self):
        from PseudoNetCDF import PseudoNetCDFFile
        mrg = self.mrg = PseudoNetCDFFile()
        mrg.createDimension('TSTEP', 9)
        mrg.variables = dict(EMIS_NO = PseudoNetCDFVariable(mrg, 'EMIS', 'f', ('TSTEP'), values = arange(9), units = 'ppb'),
                             CHEM_NO = PseudoNetCDFVariable(mrg, 'CHEM', 'f', ('TSTEP'), values = arange(9)),
                             EMIS_NO2 = PseudoNetCDFVariable(mrg, 'EMIS', 'f', ('TSTEP'), values = arange(9), units = 'ppb'),
                             CHEM_NO2 = PseudoNetCDFVariable(mrg, 'CHEM', 'f', ('TSTEP'), values = arange(9), units = 'ppb')
                            )
        self.processes = {}
        self.species = dict(NO = Species('NO'),
                            NO2 = Species('NO2'),
                            )
        exec('NOx = NO + NO2', None, self.species)
        self.testProcessFromCMAQ()

    def testProcessFromCMAQ(self):
        self.processes.update(Processes_ProcDelimSpcDict('EMIS CHEM'.split(), self.mrg.variables))
        self.assertTrue(set(self.processes.keys()) == set(['EMIS', 'CHEM']))
    
    def testIndex(self):
        testproc = self.processes['EMIS']
        NO = self.species['NO']
        NO2 = self.species['NO2']
        NOx = self.species['NOx']
        self.assertTrue((testproc[NO] == arange(9)).all())
        self.assertTrue((testproc[NOx] == (2 * arange(9))).all())
        self.assertTrue((testproc[:3][NO] == arange(9)[:3]).all())
        self.assertTrue((testproc[NO][:3] == arange(9)[:3]).all())

    def testAdd(self):
        NO = self.species['NO']
        NO2 = self.species['NO2']
        NOx = self.species['NOx']
        testproc1 = self.processes['EMIS']
        testproc2 = self.processes['CHEM']
        testproc3 = testproc1 + testproc2
        self.assertTrue((testproc3[NO] == (arange(9) * 2)).all())
        self.assertTrue((testproc3[NO2] == (arange(9) * 2)).all())
        self.assertTrue((testproc3[NOx] == (arange(9) * 4)).all())

    def testSum(self):
        NO = self.species['NO']
        NO2 = self.species['NO2']
        NOx = self.species['NOx']
        testproc1 = self.processes['EMIS']
        testproc2 = self.processes['CHEM']
        testproc3 = testproc1 + testproc2
        self.assertTrue(testproc2.sum()[NO] == arange(9).sum())
        self.assertTrue(testproc2[NO].sum() == arange(9).sum())
        self.assertTrue((testproc3.sum()[NO2] == (2 * arange(9).sum())))
        self.assertTrue(testproc3[NOx].sum() == (arange(9) * 4).sum())
    
    def testUnits(self):
        NO = self.species['NO']
        self.assertTrue(self.processes['EMIS'].get_units(NO) == 'ppb')
        self.assertTrue(self.processes['CHEM'].get_units(NO) == 'Unknown')

if __name__ == '__main__':
    unittest.main()


def PtbTable(mech):
    o3 = mech('FCONC[O3].array()')
    peak_idx = (o3 == o3.max()).nonzero()
    template = '%-8s, %-15s\t%8.1f\t%8.1f\n'
    result = ''
    
    ## Nitrogen
    #Initial NOx
    spc = 'NOx'
    proc = 'Initial'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly[0]
    peak = hourly[peak_idx]
    result += template % (spc, proc.lower(), daily, peak)
    
    #Avg NOx
    spc = 'NOx'
    proc = 'average'
    hourly = mech('(Initial[%s].array()+Final[%s].array())/2' % (spc,spc))
    daily = hourly.mean()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #Avg NOy
    spc = 'NOy'
    proc = 'average'
    hourly = mech('Initial[%s].array()' % (spc,))
    daily = hourly.mean()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #Emis NOx
    spc = 'NOx'
    proc = 'Emissions'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly.sum()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #H_Trans NOy
    spc = 'NOy'
    proc = 'H_Trans'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly.sum()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #H_Trans NOy
    spc = 'NOy'
    proc = 'V_Trans'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly.sum()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #Entrain NOy
    spc = 'NOy'
    proc = 'Entrain'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly.sum()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #Deposit NOy
    spc = 'NOy'
    proc = 'Deposit'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly.sum()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #Final NOx
    spc = 'NOx'
    proc = 'Final'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly[-1]
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #Final NOy
    spc = 'NOy'
    proc = 'Final'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly[-1]
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    ## Carbon
    #Initial VOCm
    spc = 'VOCm'
    proc = 'Initial'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly[0]
    peak = hourly[peak_idx]
    result += template % (spc, proc.lower(), daily, peak)
    
    #Avg VOCm
    spc = 'VOCm'
    proc = 'average'
    hourly = mech('(Initial[%s].array()+Final[%s].array())/2' % (spc,spc))
    daily = hourly.mean()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #Avg VOCC
    spc = 'VOCC'
    proc = 'average'
    hourly = mech('Initial[%s].array()' % (spc,))
    daily = hourly.mean()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #Emis VOCm
    spc = 'VOCm'
    proc = 'Emissions'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly.sum()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #H_Trans VOCm
    spc = 'VOCm'
    proc = 'H_Trans'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly.sum()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #H_Trans VOCm
    spc = 'VOCm'
    proc = 'V_Trans'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly.sum()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #Entrain VOCm
    spc = 'VOCm'
    proc = 'Entrain'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly.sum()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #Deposit VOCm
    spc = 'VOCm'
    proc = 'Deposit'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly.sum()
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #Final VOCm
    spc = 'VOCm'
    proc = 'Final'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly[-1]
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    #Final VOCC
    spc = 'VOCC'
    proc = 'Final'
    hourly = mech('%s[%s].array()' % (proc, spc))
    daily = hourly[-1]
    peak = hourly[peak_idx]
    result += template % (spc, proc, daily, peak)
    
    ## VOC / NOx
    #Emissions VOCm/NOx
    numerator_spc = 'VOCm'
    denominator_spc = 'NOx'
    proc = 'Emissions'
    numerator_hourly = mech('%s[%s].array()' % (proc, numerator_spc))
    numerator_daily = numerator_hourly[-1]
    numerator_peak = numerator_hourly[peak_idx]
    denominator_hourly = mech('%s[%s].array()' % (proc, denominator_spc))
    denominator_daily = denominator_hourly[-1]
    denominator_peak = denominator_hourly[peak_idx]
    daily = numerator_daily.sum()/denominator_daily.sum()
    peak = numerator_peak/denominator_peak
    result += template % ('VOCm/NOx', proc, daily, peak)
    
    #Emissions VOCm/NOx
    numerator_spc = 'VOCm'
    denominator_spc = 'NOx'
    proc = 'average'
    numerator_hourly = mech('(Initial[%s].array()+Final[%s].array())/2' % (numerator_spc, numerator_spc))
    numerator_daily = numerator_hourly[-1]
    numerator_peak = numerator_hourly[peak_idx]
    denominator_hourly = mech('(Initial[%s].array()+Final[%s].array())/2' % (denominator_spc, denominator_spc))
    denominator_daily = denominator_hourly[-1]
    denominator_peak = denominator_hourly[peak_idx]
    daily = numerator_daily.sum()/denominator_daily.sum()
    peak = numerator_peak/denominator_peak
    result += template % ('VOCm/NOx', proc, daily, peak)
    
    return result



def parse_and_run():
    import os
    from argparse import ArgumentParser
    from warnings import warn
    from permm import mechanism_dict, Mechanism
    from permm.analyses import __all__ as all_analyses
    all_mechs = '|'.join(mechanism_dict.keys())
    all_analyses = '|'.join(all_analyses)
    from PseudoNetCDF.pncparse import getparser, pncparse
    parser = getparser(has_ofile = False, interactive = True)
    parser.description =  "permm (Python Environment for Reaction Mechanism Mathematics)"
    parser.add_argument("--gui", dest="graphical", \
                        action="store_true", default=False, \
                        help="open a graphical user interactive environment")
    
    parser.add_argument("--mechanism", dest="mechanism", \
                      default="cb05_camx", help="Chemical mechanisms (e.g., a custom path|%s)" % all_mechs)

    parser.add_argument("--analysis", dest="analysis", \
                      default=None, help="Stock analysis to perform (i.e., %s)" % all_analyses)
    parser.add_argument("--scripts", dest = "scripts", action = 'append', default = [], help = "Provide scripts to run")
    parser.add_argument("--input-role", dest = "inrole", action = 'append', default = [], help = "Is this a 'merge' file or a 'concentration' file or any other word indicating type?")
    ifiles, options = pncparse(has_ofile = False, interactive = True, parser = parser)

    from permm import netcdf
    from permm.Shell import load_environ
    try:
        mech = mechanism_dict[options.mechanism]
    except KeyError:
        mech = Mechanism(options.mechanism)
        
    start_script = 0
    options.inrole += ['merge'] * (len(ifiles) - len(options.inrole))
    for r, ifile in zip(options.inrole, ifiles):
        if r == 'merge':
            mech.set_mrg(ifile)
        else:
            vardict = dict([(k, v) for k, v in ifile.variables.items() if k in mech.species_dict])
            mech.set_process(r, vardict)
            
    from permm.Shell import PERMConsole
    console = PERMConsole()
    load_environ(mech,console.locals)

    for script in options.scripts:
        if os.path.isfile(script):
            execfile(script, globals(), console.locals)
        else:
            exec(script, globals(), console.locals)


    if options.graphical:
        console.runsource("from permm.GUI import StartGUI")
        console.runsource("StartGUI(mech)")

    if options.analysis is not None:
        if len(args)<1:
            parser.error(msg="Requires a pyPA mrg file as an argument for analysis output.  You can enter an interactive environment (-i), but will not have access to \"netted\" reactions")

        if options.analysis == "net_balance":
            console.runsource("from permm.analyses.net_balance import net_balance")
            console.runsource("net_balance('%s', '%s', '%s')" % (options.mechanism, args[0], options.output))
        elif options.analysis == "history":
            console.runsource("from permm.analyses.history import matrix")
            console.runsource("history = matrix(mech, [C2O3], [HC+Radical-OH-HO2-O1D], [])")
            console.runsource("history.run()")
        else:
            raise "Unkown analysis"

    if options.interactive:
        load_environ(mech,console.locals)
        console.interact()

        
if __name__ == '__main__':
    parse_and_run()


__all__ = ['PhyTable']
from numpy import where
import warnings
from permm.core.Species import Species
warnings.formatwarning = lambda message, category, filename, lineno, file=None, line=None: '%s: %s\n' % ('ProcessWarning', message)


default_procs = ['Initial', 'Emissions', 'Chemistry', 'H_Trans', 'V_Trans', 'Entrain', 'Deposit', 'TEMPADJ', 'Final', 'Proc_Sum']
default_proc_split = ['Chemistry', 'H_Trans', 'V_Trans', 'Entrain']

def PhyTables(mech):
    result = ''
    for spck, spcv in mech.species_dict.iteritems():
        if isinstance(spcv, Species):
            result += PhyTable(mech, spck)
            result += '\n'
    
    return result
    
def VOCTable(mech):
    result = PhyTable(mech, 'VOC', speciated = True)
    result += PhyTable(mech, 'VOCC', speciated = True)
    return result
    
def PhyTable(mech, spc, extended = True, process_list = default_procs, split_list = default_proc_split, speciated = False, initial = 'Initial'):
    tag = ('', '  --XT')[extended]
    
    result = '%s%s\n' % (spc, tag)
    times = tuple(mech.mrg.variables['TFLAG'][:,0,1].tolist())
    ntimes = len(times)
    result += ('  %-19s'+"%11.4i"*ntimes+'%11s\n') % (('Time',)+times+('Sum',))
    
    if speciated:
        spc_grp = eval(spc, None, mech.species_dict)
        spcs = spc_grp.names()+[spc]
    else:
        spcs = [spc]
        
    for process in process_list:
        if speciated:
            result += '  %s\n' % process
            
        for spc in spcs:
            if speciated:
                label = "  %s" % (spc,)
            else:
                label = process
            
            if process == initial:
                result += display(mech, spc, process, label, lambda x: x, lambda x: x[0])
            else:
                result += display(mech, spc, process, label, lambda x: x)

        if extended and process in split_list:
            if speciated:
                result += '   Gains\n'

            for spc in spcs:
                if speciated:
                    label = "  %s" % (spc,)
                else:
                    if extended:
                        label = '%s, gain' % (process,)
                    else:
                        label = process
                        
                result += display(mech, spc, process, label, lambda x: where(x > 0, x, 0))

            if speciated:
                result += '   Losses\n'
            
            for spc in spcs:
                if speciated:
                    label = "  %s" % (spc,)
                else:
                    if extended:
                        label = '%s, loss' % (process,)
                    else:
                        label = process
                        
                result += display(mech, spc, process, label, lambda x: where(x < 0, x, 0))

    return result
    
def display(mech, spc, process, label, condition = None, agg = lambda x: x.sum(), chemdetail = False):
    if condition is None:
        condition = lambda x: x
    try:
        hourly = condition(mech("%s[%s].array()" % (process, spc)))
        daily = hourly.sum()
        n_vals = hourly.size+1
        values = hourly

        display_values = (label,)+tuple(values.tolist())+(agg(values),)
        result = ("  %-19s"+"%11.4f"*n_vals+'\n') % display_values
    except KeyError, strerror:
        result = ''
        warnings.warn('%s' % (strerror))

    return result


from copy import deepcopy
import re
import yaml
from numpy import float32, float64, int8, int16, int32, int64

_spc_def_re = re.compile(r'(?P<stoic>[-+]?[0-9]*\.?[0-9]+)(?P<atom>\S+)(?=\s*\+\s*)?')

def atom_parse(spc_def):
    return dict([(atom, eval(stoic)) for stoic, atom in _spc_def_re.findall(spc_def)])

def atom_guess(spc_name):
    from permm.mechanisms import atoms as ALL_ATOMS
    if '+' in spc_name or spc_name[:1].isdigit(): return atom_parse(spc_name)
    lastl = ''
    atom_dict = {}
    for l in spc_name:
        if l.isdigit():
            atom_dict[lastl] += int(l) - 1
        elif l in ALL_ATOMS:
            atom_dict.setdefault(l, 0)
            atom_dict[l] += 1
            
        lastl = l
    return atom_dict
    
class Species(object):
    def __init__(self, spc_dict, name = None, exclude = False):
        if isinstance(spc_dict, (str, unicode)):
            if not ':' in spc_dict:
                spc_dict = spc_dict.strip() + ':'
            defs = yaml.load(spc_dict)
            
            if len(defs) > 1:
                raise ValueError('Species class can only initialize one object at a time')
            else:
                (k, v), = defs.items()
                if k is False: k = 'NO'
                name = k
                if v in (None, '', 'GUESS'):
                    atom_dict = atom_guess(k)
                elif v != 'IGNORE':
                    atom_dict = atom_guess(v)
                else:
                    atom_dict = {}
                spc_dict = {k: dict(stoic = 1, atoms = atom_dict)}
        
        self.spc_dict = deepcopy(spc_dict)
        if name:
            self.name = name
        else:
            sep = {False: '+', True: '-'}[exclude]
            prefix = {False: '', True: '-'}[exclude]
            self.name = prefix + sep.join(spc_dict.keys())
            
        for spc, props in self.spc_dict.iteritems():
            props.setdefault('stoic', 1)
            props['role'] = set(props.get('role', 'rup'))
            props.setdefault('atoms', {})

        self.exclude = exclude
    
    def __getitem__(self, spc_key):
        if isinstance(spc_key, str):
            test_spc = Species({spc_key: dict(stoic = 1, role = set(['r', 'p', 'u']), atoms = {})}, name = spc_key)
            return self[test_spc]
        out_spc = {}
        for this_spc, this_props in spc_key.spc_dict.iteritems():
            if this_spc in self.spc_dict:
                check_props = self.spc_dict[this_spc]
                if this_props['role'].issubset(check_props['role']):
                    out_spc[this_spc] = new_props = {}
                    new_props['stoic'] = this_props['stoic'] * check_props['stoic']
                    new_props['atoms'] = deepcopy(this_props['atoms'])
                    new_props['atoms'].update(check_props['atoms'])
                    new_props['role'] = this_props['role']
        if len(out_spc) == 0:
            raise KeyError('%s is not in %s' % (spc_key, self))

        return Species(out_spc, exclude = spc_key.exclude)
    
    def __neg__(self):
        return Species(deepcopy(self.spc_dict), name = '-(%s)' % self.name, exclude = True)
    
    def __str__(self):
        if len(self.spc_dict) == 1:
            (k, v), = self.spc_dict.items()
            if k == self.name and v['stoic'] == 1.:
                return ('%s(%s)' % (self.name, ''.join(self.spc_dict[self.name]['role']))).replace('(pru)', '')
        bool_op = (' = ', ' != ')[self.exclude]
        result = self.name + bool_op + ' + '.join([('%.3f*%s(%s)' % (props['stoic'],spc, ''.join(props['role']))).replace('(pru)', '') for spc, props in self.spc_dict.iteritems()])
        return result
        
    def __repr__(self):
        return self.__str__()
        #try:
        #    exclude = str(self.exclude)
        #except:
        #    exclude = '?'
        #return ndarray.__repr__(self)+'\n      exclude = '+exclude
        
    def names(self):
        return [n for n in self.spc_dict.keys()]
    
    def __contains__(self, lhs):
        if isinstance(lhs, Species):
            return len(set(lhs.names()).intersection(self.names())) > 0
        elif isinstance(lhs, str):
            return lhs in self.keys()
            
    def __rmul__(self, y):
        return self.__mul__(y)

    def __mul__(self, y):
        is_number = isinstance(y,(int,long,float, float32, float64, int8, int16, int32, int64))
        
        if is_number:
            new_props = deepcopy(self.spc_dict)
            for k, props in new_props.iteritems():
                props['stoic'] *= y
            new_name = "%s * %f" % (self.name,float(y))
            new_exclude = y <= 0
            return Species(new_props, name = new_name, exclude = new_exclude)
        else:
            raise TypeError, "Can only multiply species by reactions"

    def __sub__(self, other_species):
        return self.__add__(-other_species)
        
    def __add__(self, other_species):
        return species_sum([self, other_species])
    
    def mass(self):
        from permm.mechanisms import atoms as ALL_ATOMS
        from openbabel import OBAtom
        obatom = OBAtom()
        mass = 0.
        for spc, props in self.spc_dict.items():
            for atom, count in props['atoms'].items():
                obatom.SetAtomicNum(ALL_ATOMS[atom])
                mass += obatom.GetAtomicMass() * count
        return mass*self
        
    def atoms(self, atom):
        from permm.mechanisms import atoms as ALL_ATOMS
        if atom in ALL_ATOMS:
            out_props = {}
            for spc, props in self.spc_dict.iteritems():
                mul = props['atoms'].get(atom, 0)
                if mul > 0:
                    out_props[spc] = deepcopy(props)
                    out_props[spc]['stoic'] *= mul
            if out_props == {}:
                raise KeyError, "Atom provided (%s) is not in %s" % (atom, self.name)
            else:
                return Species(out_props, name = self.name + ':' + atom, exclude = self.exclude)
        else:
            raise KeyError, "Atom provided (%s) is not an atom" % atom
    
    def copy(self):
        result =  1*self
        return result

    def stoic(self, spc = None):
        if spc is None:
            spco = self
        else:
            spco = self[spc]
        return sum(v['stoic'] for v in spco.spc_dict.itervalues())

    def iter_species_roles(self):
        for spc, props in self.spc_dict.iteritems():
            for role in props['role']:
                yield spc, role
                
    def role(self, spc = None):
        if spc is None:
            spco = self
        else:
            spco = self[spc]
        roles = [v['role'] for v in spco.spc_dict.itervalues()]

        first_role = roles[0]

        if all([first_role == role for role in roles]):
            return set(first_role)
        else:
            return set('u')

    def contains_species_role(self, spc, role):
        return role in self.spc_dict[spc]['role']
        
    def reactant(self):
        new_props = deepcopy(self.spc_dict)
        for v in new_props.itervalues():
            v['role'] = set('r')
        return Species(new_props, name = self.name, exclude = self.exclude)

    def unspecified(self):
        new_props = deepcopy(self.spc_dict)
        for v in new_props.itervalues():
            v['role'] = set('u')
        return Species(new_props, name = self.name, exclude = self.exclude)

    def product(self):
        new_props = deepcopy(self.spc_dict)
        for v in new_props.itervalues():
            v['role'] = set('p')
        return Species(new_props, name = self.name, exclude = self.exclude)

def species_sum(species_list):
    if not all([isinstance(spc,Species) for spc in species_list]):
        raise TypeError, "Can only add SpeciesGroups"
        
    include_names = set()
    exclude_names = set()
    for next_species in species_list:
        if next_species.exclude:
            exclude_names.update(next_species.spc_dict.keys())
        else:
            include_names.update(next_species.spc_dict.keys())
    
    out_include_names = include_names.difference(exclude_names)
    out_exclude_names = exclude_names.difference(include_names)
    out_props = {}
    if out_include_names != set():
        exclude = False
        out_names = out_include_names
    elif out_exclude_names != set():
        exclude = False
        out_names = out_exclude_names
    else:
        raise ValueError('Sum of species is nothing; check exclusions')
    
    for spc in species_list:
        for spcn, inprops in spc.spc_dict.iteritems():
            if not spcn in out_names: continue

            outprops = out_props.setdefault(spcn, dict(stoic = 0., role = set(), atoms = {}))
            outprops['stoic'] += inprops['stoic']
            outprops['role'].update(inprops['role'])
            outatoms = outprops['atoms']
            inatoms = inprops['atoms']
            for atom, value in inatoms.iteritems():
                try:
                    outatoms[atom] += value
                except:
                    outatoms[atom] = value

    return Species(out_props, exclude = exclude)


import unittest

class SpeciesTestCase(unittest.TestCase):
    def setUp(self):
        self.species = dict(OH = Species(dict(OH = dict(stoic = 1, atoms = dict(H = 1, O = 1))), name = 'OH', exclude = False),
                            HO2 = Species(dict(HO2 = dict(stoic = 1, atoms = dict(H = 1, O = 2))), name = 'HO2', exclude = False),
                            HOx = Species(dict(OH = dict(stoic = 1, atoms = dict(H = 1, O = 1)), HO2 = dict(stoic = 1, atoms = dict(H = 1, O = 2))), name = 'HOx', exclude = False),
                            O3 = Species(dict(O3 = dict(stoic = 1, atoms = dict(O = 3))), name = 'O3', exclude = False),
                            )


    def assertEqualSpecies(self, s1, s2, neg = False):
        self.assertEquals(s1.spc_dict, s2.spc_dict)
        if neg:
            self.assertEquals(s1.exclude, not s2.exclude)
        else:
            self.assertEquals(s1.exclude, s2.exclude)
        
    def testCreate(self):
        s1 = self.species['OH']
        s2 = Species(s1.spc_dict, name = 'hydroxyl')
        self.assertEqualSpecies(s1, s2)
        
    def testNeg(self):
        s1 = self.species['OH']
        s2 = -s1
        self.assertEqualSpecies(s1, s2, neg = True)
    
    def testAdd(self):
        s1 = self.species['OH']
        s2 = self.species['HO2']
        s3 = self.species['HOx']
        s4 = s1 + s2
        self.assertEqualSpecies(s3, s4)

    def testSub(self):
        s1 = self.species['OH']
        s2 = self.species['HO2']
        s3 = self.species['HOx']
        s4 = s3 - s2
        self.assertEqualSpecies(s4, s1)

    def testGet(self):
        s2 = self.species['HO2']
        s3 = self.species['HOx']
        self.assertEqualSpecies(s3['HO2'], s3[s2])

    def testAtoms(self):
        s2 = self.species['HO2']
        s3 = self.species['HOx']
        self.assertEquals(s2.atoms('O').stoic(s2), 2)
        self.assertEquals(s3.atoms('O').stoic(s2), 2)
        self.assertEquals(s3.atoms('O').stoic(s3), 3)

    def testMul(self):
        s1 = self.species['OH']
        s3 = 2 * s1
        s2 = s1 * 2
        self.assertEqualSpecies(s2, s3)
        self.assertRaises(AssertionError, self.assertEqualSpecies, s1, s3)
        
    def testContains(self):
        s1 = self.species['OH']
        s2 = self.species['O3']
        s3 = self.species['HOx']
        self.assertTrue(s1 in s3)
        self.assertFalse(s2 in s3)

    def testCopy(self):
        s1 = self.species['OH']
        s2 = s1.copy()
        self.assertEqualSpecies(s1, s2)

    def testNewFromSpc(self):
        s1 = self.species['OH']
        s2 = self.species['O3']
        s3 = self.species['HOx']
        s4 = Species(s1.spc_dict)
        s5 = s1 + s2 + s3
        self.assertEqualSpecies(s1, s4)
        self.assertEqualSpecies(s1 + s2 + s3, s5)

if __name__ == '__main__':
    unittest.main()



import operator
import re
from warnings import warn
from copy import deepcopy
from numpy import ndarray, \
                  newaxis, \
                  float64, \
                  vectorize, \
                  rollaxis, \
                  array, \
                  indices, \
                  equal, \
                  greater, \
                  less
from numpy.ma import sum, masked_less, masked_greater

from permm.core.Species import Species

ReactionGroup = str

__all__ = ['Stoic', 'Reaction']


class Stoic(ndarray):
    """
    Stoic is a sub-class of float with a role
    property.  The role property is 'r', 'p', or 'u'.
    r = reactant
    p = product
    u = undetermined
    
    Undetermined is useful when a species is part of a 
    net reaction and may play either role
    
    Stoic are created by
        Stoic(2., 'r')
    
    Stoic supports:
        __mul__: Stoic*(Stoic|Number)
        __rmul__: (Stoic|Number)*Stoic
        __add__: Stoic+(Stoic|Number)
    """
    __array_priority__ = 1001.
    def __new__(subtype, f, role):
        result = array(f).view(subtype)
        result.role = role
        return result
    
    def __array_wrap__(self, obj, context = None):
        result = obj.view(Stoic)
        if context is not None:
            args = context[1]
            if len(args) > 1:
                lhs, rhs = args[:2]
                if hasattr(lhs, 'role'):
                    if hasattr(rhs, 'role'):
                        if lhs.role != rhs.role:
                            result.role = 'u'
                        else:
                            result.role = lhs.role
                    else:
                            result.role = lhs.role                        
                else:
                    result.role = 'u'

        return result
        
    def __repr__(self):
        return "Stoic(%s, role = '%s')" % (str(self.asarray()), getattr(self, 'role', 'None'))
    
    def __str__(self):
        return self.__repr__()
            
    def asarray(self):
        return self.view(ndarray)
        

StoicArray = vectorize(Stoic, otypes = [object])
def AddRole(s1, s2):
    if isinstance(s1,Stoic):
        return s1
    else:
        return Stoic(s1, s2.role)
        
def StoicAdd(s1, s2):
    return Stoic.__add__(AddRole(s1,s2), s2)

StoicAdd = vectorize(StoicAdd, otypes = [object])

def ParseReactionString(rxn_str):
    """
    ReactionFromString is a convenience function.  It creates
    Reaction objects from a string with the following pattern:
        "(?P<reactants>.*)=(?P<rxn_type>[kj])[>]\s*(?P<products>.*)"
    where each reactant or product matches the following pattern:
        "(\s?(?P<sign>[+-])?\s?)?((?P<stoic>\d(\.(\d{1,3}(E\d{2})?)?)?)\*)?(?P<name>[A-Z]\w{0,3})\s?"

    For example:
        OH + OLE =k> 0.8*FORM + 0.33*ALD2 + 0.62*ALDX + 0.8*XO2 + 0.95*HO2 - 0.7 PAR
    """
    stoics = {}
    
    reaction_re = re.compile("(?P<reactants>.*)(?:=|->\[(?P<rxn_type>[kjdeu])\])\s*(?P<products>.*)")

    species_re = re.compile("(\s?(?P<sign>[+-])?\s?)?((?P<stoic>\d{0,1}(\.(\d{1,3}(E\d{2})?)?)?)\*)?(?P<name>[a-zA-Z]\w*)(?:[ +=]|$)+",re.M)
    
    reaction_match = reaction_re.match(rxn_str)
    if reaction_match is None:
        raise SyntaxError, "Reactions must match the following patter\n\n<%%d> stoic*spc ([+-] stoic*spc)* =[kjude]> [stoic*spc] ([+-] stoic*spc)*\n\n%s" % (rxn_str,)

    reactants = reaction_match.groupdict()['reactants']
    reaction_type = reaction_match.groupdict()['rxn_type']
    if reaction_type is None:
        reaction_type = 'u'
    products = reaction_match.groupdict()['products']

    for spc in species_re.finditer(reactants):
        spc_g = spc.groupdict()
        name = spc_g['name']
        sign = spc_g['sign']
        stoic = spc_g['stoic']
        if sign is None:
            sign = ''

        if stoic is None:
            stoic = '1'

        if (name, 'r') in stoics:
            stoics[name, 'r'] += -float(sign + stoic)
        else:
            stoics[name, 'r'] = -float(sign + stoic)

    
    for spc in species_re.finditer(products):
        spc_g = spc.groupdict()
        name = spc_g['name']
        sign = spc_g['sign']
        stoic = spc_g['stoic']
        if sign is None:
            sign = '+'

        if stoic is None:
            stoic = '1'
        
        value = float(sign + stoic)
        if (name, 'p') in stoics:
            stoics[name, 'p'] += value
        else:
            stoics[name, 'p'] = value

    return stoics, reaction_type
    
def ReactionFromString(rxn_str):
    return Reaction(*ParseReactionString(rxn_str))

class Reaction(object):
    """
    Reaction is an object that represents reaction groups.  The simplest
    case being a "single reaction" reaction group.
    
    Reaction groups support the following interfaces

        1) indexing (__getitem__) for stoiciometry
        2) multiplication (__mul__ and __rmul__) by arrays and numbers
        3) addition (__add__) of reactions or species
    
    A Reaction can also determine when a species is a reactant, product 
    or unspecified.  When a species is only present as a one role or the other 
    (reactant or product), it is always that role.  When a species is present 
    in the group as both, its current role is determined by the stoichiometry.
    
    Convenience functions:
        has_spc = spc in Reaction.species()
        has_rct = spc in Reaction.reactants()
        has_prd = spc in Reaction.products()
    
    Other functions:
        add_rct_spc
        add_prd_spc
    """
    __array_priority__ = 1000.

    def __init__(self, stoic, reaction_type = 'k', safe = False):
        """
            roles - dictionary of specific roles; for species whose 
                    stoichiometric sign is inconsistent with their 
                    role (i.e. X + Y =k> Z - .6*PAR)
            reaction_type - 'k' is kinetic, 'j' is photolysis, 'n' is net
            stoic - stoichiometry values provided as keywords by species; values
                    can be scalars or ndarrays
            safe - when species roles are not found, find any version of
                   that species to prevent an error
        """
        if isinstance(stoic, str):
            stoic, reaction_type = ParseReactionString(stoic)

        self.reaction_type = reaction_type
        self._safe = safe
        self._stoic = deepcopy(stoic)
        self._update_roles()
        try:
            first_key = self._stoic.keys()[0]
            self.shape = self._stoic[first_key].shape
        except AttributeError, (e):
            self.shape = ()
            for k, v in self._stoic.iteritems():
                self._stoic[k] = float64(v)
        except IndexError, (e):
            self.shape = ()
            
    
    def _update_roles(self):
        keys = self._stoic.keys()
        self._species = tuple(set([spcn for spcn, role in keys]))
        self._reactants = tuple(set([spcn for spcn, role in keys if role == 'r']))
        self._products = tuple(set([spcn for spcn, role in keys if role == 'p']))
        self._unspecified = tuple(set([spcn for spcn, role in keys if role == 'u']))
        
    def roles(self, spc):
        roles = set()
        if isinstance(spc, str):
            if spc in self._reactants:
                roles.update('r')
            if spc in self._products:
                roles.update('p')
            if spc in self._unspecified:
                roles.update('u')
                
        elif isinstance(spc, Species):
            roles.update(*[self.roles(spcn) for spcn in spc.spc_dict.iterkeys()])
        
        return roles
            
    def __contains__(self, lhs):
        """
        Test if reaction has a species
        """
        if isinstance(lhs, Species):
            return self.has_spc(lhs)
        elif isinstance(lhs, str):
            return lhs in self._species
        else:
            raise TypeError, 'Unknown comparison: __contains__ for Reaction and %s' % str(type(lhs))
            
    def __getitem__(self, item):
        """
        Return a stoichiometry for a species or, if item is not a species, return
        a new reaction where stoichiometry is subset using item
        """
        if isinstance(item, Species):
            return self.get_spc(item)
        elif isinstance(item, str):
            if  item in self._species:
                outvalue = 0.
                outrole = set()
                for role in self.roles(item):
                    outvalue += self._stoic[item, role]
                    outrole.update(role)

                if len(outrole) == 1:
                    role, = outrole
                else:
                    role = 'u'
                
            return Stoic(outvalue, role = role)
        else:
            return Reaction(dict([(k,v[item]) for k, v in self._stoic.iteritems()]), reaction_type = self.reaction_type)
    
    def __str__(self):
        """
        Report all values followed by the sum value for the reaction
        """
        result = ''
        temp = self.copy()
        temp.shape = ()
        if self.shape != ():
            result += '%d Reactions with shape %s\n' % (reduce(int.__mul__, self.shape), str(self.shape))
            for idx in indices(self.shape).reshape(len(self.shape), -1).swapaxes(0,1):
                idx = tuple(idx)
                result += str(idx) + ': '
                for spc in self._stoic.keys():
                    temp._stoic[spc] = self._stoic[spc][idx]
                temp._update_roles()
                result += str(temp)+', \n'
            result = result[:-3]+'\n'
        sum_result = self.display(digits = 5, nspc = 1000)

        if result != '':
            result += '-' * (len(sum_result)+5) + '\nSum: '
        result += sum_result
            
        return result

    def __repr__(self):
        """
        Representation is same as string
        """
        return self.__str__()

    def display(self, digits = 5, nspc = 3, formatter = 'g'):
        reactants = []
        products = []
        for (spcn, role), v in self._stoic.iteritems():
            if role == 'r':
                reactants.append((self._stoic[spcn, role].sum(), spcn))
            elif role == 'p':
                products.append((self._stoic[spcn, role].sum(), spcn))
            else:
                v = self._stoic[spcn, role].sum()
                if v < 0:
                    reactants.append((v, spcn))
                else:
                    products.append((v, spcn))

        reactants.sort(reverse=False)
        products.sort(reverse=True)

        if digits == None:
            str_temp = '%s'
            reactant_str = ' + '.join([str_temp % rct for stoic, rct in reactants][:nspc])
            product_str = ' + '.join([str_temp % prd for stoic, prd in products][:nspc])
        elif digits == -1:
            reactant_str = ' + '.join(['*'.join([str(-stoic), str(rct)]) for stoic, rct in reactants][:nspc])
            product_str = ' + '.join(['*'.join([str(stoic), str(prd)]) for stoic, prd in products][:nspc])
        else:
            str_temp = '%%.%d%s*%%s' % (digits, formatter)
            reactant_str = ' + '.join([str_temp % (-stoic,rct) for stoic, rct in reactants][:nspc])
            product_str = ' + '.join([str_temp % (stoic,prd) for stoic, prd in products][:nspc])

        if len(reactants) > nspc:
            reactant_str += ' + ...'
        if len(products) > nspc:
            product_str += ' + ...'

        return '%s ->[%s] %s' % (reactant_str, self.reaction_type, product_str)
        
    def __add__(self,rhs):
        """
        Add reactions to make a net reaction or add species to an existing reaction.
        """
        if isinstance(rhs,Reaction):
            kwds = {}
            
            for key, v in self._stoic.iteritems():
                kwds[key] = self._stoic[key].copy()
            
            for key, v in rhs._stoic.iteritems():
                if kwds.has_key(key):
                    kwds[key] += rhs._stoic[key].copy()
                else:
                    kwds[key] = rhs._stoic[key].copy()

            if self.reaction_type == rhs.reaction_type:
                reaction_type = self.reaction_type
            else:
                reaction_type = 'n'

        elif isinstance(rhs,Species):
            return self.__add_if_in_spclist(rhs,self._species)

        else:
            raise TypeError, "Currently, only reactions can be added together"
        
        return Reaction(kwds, reaction_type = reaction_type)
        
    def __rmul__(self,irrs):
        return self.__mul__(irrs)
        
    def __mul__(self,irrs):
        species = self._species
        values = dict([(k,v*irrs) for k, v in self._stoic.iteritems()])
        result = Reaction(values, reaction_type = self.reaction_type)
        return result

    def __add_if_in_spclist(self,rhs,spc_list):
        if not self.has_spc(rhs):
            raise KeyError, 'Reaction has no components of %s' % rhs.name
        elif rhs.name in self._species:
            warn('Already has %s' % rhs.name)
            return self.copy()
        elif rhs.exclude:
            raise ValueError, 'Exclude is not supported'
        result = self.copy()
        new_stoic = result[rhs]
        
        result._roles[rhs.name] = new_stoic.role
        result._species += (rhs.name,)
        result._stoic[rhs.name] = new_stoic.view(ndarray)
        result._update_roles()

        return result
            
    def copy(self):
        """
        Create a copy of the reaction such that stoichiometry 
        are not shared
        """
        return Reaction(dict([(k, v.copy()) for k, v in self._stoic.iteritems()]), reaction_type = self.reaction_type, )

    def sum(self, axis = None):
        """
        Sum stoichiometries and create a scalar reaction
        """
        result = self.copy()
        for key in result._stoic.keys():
            result._stoic[key] = self._stoic[key].sum(axis)
        result.shape = ()
        result._update_roles()
        return result
        
    def mean(self, axis = None):
        """
        Mean stoichiometries and create a scalar reaction
        """
        result = self.copy()
        for spc in result._species:
            result._stoic[spc] = self._stoic[spc].mean(axis)
        result.shape = ()
        result._update_roles()
        return result
        
    def reactants(self):
        """
        Report all species acting as reactants; includes negative
        stoichiometries for unspecified role species
        """
        return self._reactants
        
    def products(self):
        """
        Report all species acting as products; includes positive
        stoichiometries for unspecified role species
        """
        return self._products
        
    def species(self):
        """
        Report all species
        """
        return self._species

    def unspecified(self):
        """
        Report all species with unspecified roles
        """
        return self._unspecified

    def get(self, item, default = None):
        try:
            return self.__getitem__(item)
        except:
            return default

    def get_spc(self, *args):
        nargs = len(args)
        if nargs > 2:
            raise TypeError('get_spc expected at most 2 arguments, got %d' % nargs)
            
        item = args[0]

        values = []
        roles = []
        if item.exclude:
            item_spc_roles = [k for k in item.iter_species_roles()]
            item_spcs = [k[0] for k in item_spc_roles]
            for spc, role in self._stoic:
                if role == 'u' and spc in item_spcs and (spc, 'u') not in item_spc_roles:
                    val = self._stoic[spc, role]
                    if greater(val, 0).all():
                        role = 'p'
                    elif less(val, 0).all():
                        role = 'r'
                
                if not (spc, role) in item_spc_roles:
                    values.append(1 * self._stoic[spc, role])
                    roles.append(role)
                    
                
        else:
            for spc, props in item.spc_dict.iteritems():
                spc_roles = props['role']
                for role in spc_roles:
                    if (spc, role) in self._stoic:
                        values.append(item.spc_dict[spc]['stoic'] * self._stoic[spc, role])
                        roles.append(role)
                
                if 'u' not in spc_roles and (spc, 'u') in self._stoic:
                    val = self._stoic[spc, 'u']
                    if role == 'p':
                        values.append(item.spc_dict[spc]['stoic'] * masked_less(val, 0).filled(0))
                        roles.append(role)
                    elif role == 'r':
                        values.append(item.spc_dict[spc]['stoic'] * masked_greater(val, 0).filled(0))
                        roles.append(role)
                    
                
                
        if len(values) == 0:
            if self._safe:
                if nargs == 1:
                    if item.name in self:
                        warn("%s not in %s; trying all roles" % (item, self.sum()))
                        return self.get_spc(Species(item.name, exclude = item.exclude))
                    raise KeyError, "%s does not contain %s" % (str(self.sum()), str(item))
                else:
                    return args[1]
            else:
                raise KeyError, "%s does not contain %s" % (str(self.sum()), str(item))
            
                
        last_role = roles[-1]
        same_role = all([last_role == role for role in roles])
        
        if same_role:
            role = last_role
        else:
            role = 'u'
        return Stoic(sum(values, axis = 0), role = role)
    
    def produces(self, item):
        try:
            return self.get_spc(item.product())
        except KeyError:
            return 0 * self._stoic[self._stoic.keys()[0]]

    def consumes(self, item):
        try:
            return -self.get_spc(item.reactant())
        except KeyError:
            return 0 * self._stoic[self._stoic.keys()[0]]
    
    def has_spc(self,spc_grp):
        for spc_role in spc_grp.iter_species_roles():
            if spc_role in self._stoic:
                return True != spc_grp.exclude
        return False != spc_grp.exclude

    def has_rct(self,spc_grp):
        for name in spc_grp.names():
            if not spc_grp.contains_species_role(name, 'r'):
                raise TypeError('Requesting reactant role from species %s with %s that has roles %s' % (spc_grp.name, name, str(list(spc_grp.spc_dict[name]['role']))))
        return self.has_spc(spc_grp.reactant())
        
    def has_prd(self,spc_grp):
        for name in spc_grp.names():
            if not spc_grp.contains_species_role(name, 'p'):
                raise TypeError('Requesting product role from species %s with %s that has roles %s' % (spc_grp.name, name, str(list(spc_grp.spc_dict[name]['role']))))
        return self.has_spc(spc_grp.product())
    
    def has_unspc(self,spc_grp):
        for name in spc_grp.names():
            if not spc_grp.contains_species_role(name, 'u'):
                raise TypeError('Requesting unspecified role from species %s with %s that has roles %s' % (spc_grp.name, name, str(list(spc_grp.spc_dict[name]['role']))))
        return self.has_spc(spc_grp.unspecified())
    
    def net(self, spco = None):
        if spco is None:
            spcs_to_condense = self._species
        else:
            if spco.name in self:
                spcs_to_condense = [spco.name]
            else:
                spcs_to_condense = spco.spc_dict.keys()
        net_spcs = []
        for spc in spcs_to_condense:
            prd = spc in self._products
            rct = spc in self._reactants
            unk = spc in self._unspecified
            if not prd ^ rct ^ unk:
                net_spcs.append(spc)
        kwds = deepcopy(self._stoic)
        dummy_stoic = Stoic(0., role = 'u')
        for spc in net_spcs:
            rval = kwds.pop((spc, 'r'), dummy_stoic)
            pval = kwds.pop((spc, 'p'), dummy_stoic)
            uval = kwds.pop((spc, 'u'), dummy_stoic)
            new_val = rval + pval + uval

            if not equal(new_val, 0.).all():
                kwds[spc, 'u'] = new_val

        return Reaction(kwds, reaction_type = self.reaction_type)

    def condense(self, spco, name = None):
        if not isinstance(spco, Species):
            raise TypeError('condense can only take Species objects')
        if spco in self:
            net_keys = [(spc, role) for spc, role in spco.iter_species_roles() if (spc, role) in self._stoic]
            new_roles = set([role for spc, role in net_keys])
            new_val_dict = dict([(role, []) for role in new_roles])
            new_name = name or spco.name
            
            kwds = deepcopy(self._stoic)
            for spc, role in net_keys:
                new_val_dict[role].append(kwds.pop((spc, role)))

            for new_role, new_vals in new_val_dict.iteritems():
                kwds[new_name, new_role] = reduce(operator.add, new_vals)

            return Reaction(kwds, reaction_type = self.reaction_type)
        else:
            return self.copy()
        
import unittest

class ReactionTestCase(unittest.TestCase):
    def setUp(self):
        self.spcs = dict(NO2 = Species('NO2'),
                         NO = Species('"NO"'),
                         HNO3 = Species('HNO3'),
                         O = Species('O'),
                         OH = Species('OH'),
                         HO2 = Species('HO2'),
                         O3 = Species('O3'),
                         PAR = Species('PAR: C'),
                         NTR = Species('NTR: IGNORE'),
                         FORM = Species('FORM: CH2O'),
                         ALD2 = Species('ALD2: CH3CHO'),
                         ALDX = Species('ALDX: CH3CH2CHO'),
                         )
        exec('NOx = NO2 + NO', globals(), self.spcs)
        exec('ALD = ALDX + ALD2 + FORM', globals(), self.spcs)
        rxn_strings = {'NO2hv': 'NO2 ->[j] NO + O',
                       'OplO3': 'O + O2 + M ->[k] O3 + M',
                       'NTRplOH': 'NTR + OH ->[k] HNO3 + HO2 + 0.330*FORM + 0.330*ALD2 + 0.330*ALDX - 0.660*PAR'
                      }
        self.rxns = {}
        for label, rxn_str in rxn_strings.iteritems():
            self.rxns[label] = Reaction(rxn_str)
        
    def testHasSpc(self):
        r1 = self.rxns['NO2hv']
        spcs = [self.spcs[k] for k in ['NO2', 'NO', 'NOx']]
        for spc in spcs:
            self.assertTrue(r1.has_spc(spc))
            
    def testHasRole(self):
        r1 = self.rxns['NO2hv']
        rs = self.spcs['NO2']
        ps = self.spcs['NO']
        self.assertTrue(r1.has_rct(rs))
        self.assertFalse(r1.has_rct(ps))
        self.assertFalse(r1.has_prd(rs))
        self.assertTrue(r1.has_prd(ps))
        
    def testConsumes(self):
        r1 = self.rxns['NO2hv']
        rs = self.spcs['NO2']
        ps = self.spcs['NO']
        ns = self.spcs['O3']
        self.assertTrue(r1.consumes(rs) == 1)
        self.assertTrue(r1.consumes(ps) == 0)
        self.assertTrue(r1.consumes(ns) == 0)
        
    def testProduces(self):
        r1 = self.rxns['NO2hv']
        rs = self.spcs['NO2']
        ps = self.spcs['NO']
        ns = self.spcs['O3']
        self.assertTrue(r1.produces(rs) == 0)
        self.assertTrue(r1.produces(ps) == 1)
        self.assertTrue(r1.produces(ns) == 0)
        
    def testGet(self):
        r1 = self.rxns['NO2hv']
        rs = self.spcs['NO2']
        self.assertTrue(r1[rs] == -1)
        
    def testReactants(self):
        r1 = self.rxns['NTRplOH']
        self.assertEquals(set(r1.reactants()), set(('OH', 'NTR')))
        
        
    def testProducts(self):
        r1 = self.rxns['NTRplOH']
        self.assertEquals(set(r1.products()), set(('PAR', 'FORM', 'ALD2', 'HNO3', 'HO2', 'ALDX')))
        
    def testMulScalar(self):
        r1 = self.rxns['NTRplOH']
        a = 5.83
        r2 = r1 * a

        for spcn in "NTR OH".split():
            spc = self.spcs[spcn]
            self.assertAlmostEqual(r2[spc], -a)

        for spcn in "HNO3 HO2".split():
            spc = self.spcs[spcn]
            self.assertAlmostEqual(r2[spc], a)

        for spcn in "FORM ALD2 ALDX".split():
            spc = self.spcs[spcn]
            self.assertAlmostEqual(r2[spc], a * .33)

        for spcn in ["ALD"]:
            spc = self.spcs[spcn]
            self.assertAlmostEqual(r2[spc], a * .99)

        for spcn in ["PAR"]:
            spc = self.spcs[spcn]
            self.assertAlmostEqual(r2[spc], -a * .66)
        
    def testMulArray(self):
        from numpy import arange, round
        a = arange(0, 60, dtype = 'd').reshape(3,4,5) + .3
        r1 = self.rxns['NTRplOH']
        r2 = r1 * a

        for spcn in "NTR OH".split():
            spc = self.spcs[spcn]
            self.assertTrue((r2[spc] == -a).all())

        for spcn in "HNO3 HO2".split():
            spc = self.spcs[spcn]
            self.assertTrue((r2[spc] == a).all())

        for spcn in "FORM ALD2 ALDX".split():
            spc = self.spcs[spcn]
            self.assertTrue((r2[spc] == (a * float64(.33))).all())

        for spcn in ["ALD"]:
            spc = self.spcs[spcn]
            # Value is not exact because of addition
            self.assertTrue((round(r2[spc], decimals = 8) == round(a * float64(.99), decimals = 8)).all())

        for spcn in ["PAR"]:
            spc = self.spcs[spcn]
            self.assertTrue((r2[spc] == (-a * .66)).all())
        
if __name__ == '__main__':
    unittest.main()
    r1 = Reaction('NO2 ->[j] NO + O')
    r2 = Reaction('O + O2 + M ->[k] O3 + M')
    NO2 = Species('NO2')
    NO = Species('NO')
    O = Species('O')
    O3 = Species('O3')
    Ox = O3 + NO2
    Ox.name = 'Ox'
    print r1[NO2]
    print r1[NO]
    print r1[O]
    from numpy import arange
    nr = (r1 * arange(9) * 1.1 + r2 * arange(9)[::-1]).net()
    print nr[:4][O.reactant()]
    print nr[O]
    try:
        print nr[O.reactant()]
    except:
        pass
    print nr.net()
    print nr.net().condense(Ox)


import re
import sys
import yaml

from Mechanism import Mechanism

reaction_pattern = re.compile("^(?P<r_num>\d{1,3})\s(?P<reactants>[A-Z]\w{0,3}(\s\+\s[A-Z]\w{0,3})*)\s\n?(?P<products>(\d(\.(\d{1,3}(E\d{2})?)?)?\s)?[A-Z]\w{0,3}(\s\n?(\+|\-)\n?\s\n?(\d(\.(\d{1,3}(E\d{2})?)?)?\s\n?)?[A-Z]\w{0,3})*)?\s\n?(?P<reaction_type>Photolysis|(\d\.\d{3}E[+-]\d{2}))",re.M)

species_pattern = re.compile("(\s?\n?(?P<sign>[+-])?\s?\n?)?(?P<stoic>\d(\.(\d{1,3}(E\d{2})?)?)?\s\n?)?(?P<name>[A-Z]\w{0,3})\s?",re.M)
reactions_text = file("../../mechanisms/cb05_camx/cb05_camx_pdf.txt","r").read()

reaction_list = [match.groupdict() for match in reaction_pattern.finditer(reactions_text)]

tab = "    "
exclude_species = ("O2", "M", "H2O", "None")
species_list = []
reaction_yaml = "reaction_list:\n"
species_yaml = "species_list:\n"

def consolidate_dups(spc_group):
    spc_group_names = set([s['name'] for s in spc_group])
    if len(spc_group_names) < len(spc_group):
        new_spc_group=[dict(name=name) for name in spc_group_names]
        for new_spc in new_spc_group:
            for old_spc in [old_spc for old_spc in spc_group if old_spc['name'] == new_spc['name']]:
                if new_spc.get('sign',old_spc['sign']) != old_spc['sign']:
                    raise ValueError, 'Math not implemented for different signs'
                else:
                    new_spc['stoic'] = str(float(new_spc.get('stoic',0))+float(old_spc['stoic']))
                    new_spc['sign'] = old_spc['sign']
        return new_spc_group
    else:
        return spc_group

def remove_nones(spc_group):
    spc_group_names = [s['name'] for s in spc_group]
    while None in spc_group_names:
        spc_group.pop(spc_group_names.index(None))
        spc_group_names.pop(spc_group_names.index(None))
    return spc_group

def set_reactant_defaults(reactants):
    for reactant in reactants:
        reactant['sign'] = '-'
        if reactant['stoic'] is None:
            reactant['stoic'] = '1'
    return consolidate_dups(remove_nones(reactants))

def set_product_defaults(products):
    for product in products:
        if product['sign'] is None:
            product['sign'] = '+'
        if product['stoic'] is None:
            product['stoic'] = '1'
    product_names = [p['name'] for p in products]
    return consolidate_dups(remove_nones(products))

p = re.compile('(\d) ([A-Z])')

for ri,rxn in enumerate(reaction_list):
    reactants = p.sub(r'\1*\2',rxn["reactants"])

    products = p.sub(r'\1*\2',str(rxn["products"]).replace('None','').replace('\n',''))
    
    if products is None:
        products = ''
    
    if rxn['reaction_type'] == 'Photolysis':
        reaction_type = 'j'
    else:
        reaction_type = 'k'
    
    reaction_yaml+=(1*tab)+("RXN_%02d: %s =%s> %s" % (int(rxn["r_num"]),reactants,reaction_type,products)).replace('\n','').replace('  ',' ')+'\n'
        
    

        
    reactants = set_reactant_defaults( \
                  [i.groupdict() for i in \
                       species_pattern.finditer(reactants) \
                  ])

    products =  set_product_defaults( \
                  [i.groupdict() for i in \
                       species_pattern.finditer(products)
                  ])

    for role,spc_group in [("r", reactants), ("p", products)]: 
        for spc in spc_group: 
            spc["name"] = spc["name"].replace("\n","")
            if spc["name"] not in exclude_species:
                if spc["name"] not in species_list:
                    species_list.append(spc["name"])
                    species_yaml += 1 * tab + "- &%s '%s'\n" % (spc["name"],spc["name"])


initial_yaml = '---\n' + \
               species_yaml + '\n' +  \
               reaction_yaml + '\n' + \
               file('../../mechanisms/cb05_camx/cb05_camx_new_groups.yaml').read() + '\n...'

print >> file('../../mechanisms/cb05_camx/cb05_camx.yaml','wb'), initial_yaml
initial_mech = Mechanism(initial_yaml)
globals().update(initial_mech.species_dict)
net_reaction_rules=yaml.load(file('../../rules.yaml'))
net_reaction_yaml='\n'
net_reaction_yaml='net_reaction_list:\n'

def get_species(spc_str):
    return eval('['+spc_str+']',initial_mech.species_dict)
    
for net_key, net_rule in net_reaction_rules.iteritems():
    rxns = net_rule.get('rxns', [])
    if net_rule.has_key('r') or net_rule.has_key('p') or not net_rule.has_key('rxns'):
        reactants = get_species(net_rule.get('r',''))
        products = get_species(net_rule.get('p',''))
        logical_and = net_rule.get('logical_and',True)
        rxns += initial_mech.find_rxns(reactants,products,logical_and)
    
    
    net_reaction_yaml += '    %s: %s\n' % (net_key, ' + '.join(rxns))

full_yaml = '---\n' + \
               species_yaml + '\n' +  \
               reaction_yaml + '\n' + \
               net_reaction_yaml + '\n' + \
               file('../../mechanisms/cb05_camx/cb05_camx_new_groups.yaml').read() + '\n' + \
               file('camx_netphysical26.yaml').read() + '\n...'
               
# + '\n' + \
#               file('diagram.yaml').read()
print full_yaml


__all__ = ['AttrDict']

import unittest

class AttrDict(dict):
    """For easy access to values"""
    def __new__(cls,*args,**kwds):
        result = dict.__new__(cls,*args,**kwds)
        return result
        
    def __init__(self,*args,**kwds):
        dict.__init__(self, *args, **kwds)
        for k,v in self.iteritems():
            if type(v)==dict:
                self[k]=AttrDict(v)
            else:
                self[k]=v
                
    def __setattr__(self, attr, value):
        self[attr] = value
        
    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(attr)


class TestAttrDict(unittest.TestCase):
    def setUp(self):
        self.base_dict = {'one': 1, "two": {"two": 2}}
        self.attr_dict = AttrDict(self.base_dict)
        
    def testInitFromDict(self):
        bd = self.base_dict
        ad = self.attr_dict
        for k in self.base_dict:
            self._assert(bd[k] == ad[k])
            


__all__ = ['IRRTable', 'NetTableMaker', 'PhyTableMaker', 'PtbMaker', 'SumTableMaker']

import IRRTable
import NetTableMaker
import PhyTableMaker
import PtbMaker
import SumTableMaker

def net_balance(mechanism, mrg_data_path, output_dir):
    from permm.analyses.net_balance.PhyTableMaker import PhyTable
    from permm.analyses.net_balance.SumTableMaker import SumTable
    from permm.analyses.net_balance.NetTableMaker import NetTables
    from permm.analyses.net_balance.PhyTableMaker import PhyTables
    from permm.analyses.net_balance.PhyTableMaker import VOCTable
    from permm.analyses.net_balance.PtbMaker import PtbTable
    from permm.analyses.net_balance.IRRTable import IRRTable
    from permm import getmech
    from permm.netcdf import NetCDFFile
    
    get_prepared_mech = getmech.get_prepared_mech
    get_pure_mech = getmech.get_pure_mech
    try:
        mech = get_prepared_mech(mechanism)
    except:
        mech = get_pure_mech(mechanism)
    

    if output_dir is None:
        net_data_path = '.'.join(mrg_data_path.split('.')[:-1])+'.net.'
    else:
        net_data_path = output_dir+'.'
    
    mrg_file = NetCDFFile(mrg_data_path, 'r')
    mech.set_mrg(mrg_file)

    print >> file(net_data_path+'sum','wb'), SumTable(mech)
    print >> file(net_data_path+'net','wb'), NetTables(mech)
    print >> file(net_data_path+'phy','wb'), PhyTables(mech)
    print >> file(net_data_path+'voc','wb'), VOCTable(mech)
    print >> file(net_data_path+'ptb','wb'), PtbTable(mech)
    mech = get_pure_mech(mechanism)
    mech.set_mrg(mrg_file)

    
    print >> file(net_data_path+'irr','wb'), IRRTable(mech)


__all__ = ['profile_process']

from warnings import warn
from pylab import figure, title, bar, savefig, legend, grid, xlabel, ylabel, xticks, axis
from matplotlib.cm import get_cmap
from matplotlib.colors import rgb2hex
from numpy import arange, ndarray

def profile_process(mech, species_group, process, loc = 1, width = .8, sort = False, fig = None, cmap = None, normalize = True, missing = None):

    # If necessary, start a new figure.  Otherwise,
    # store the previous colors as a dictionary
    if fig is None:
        fig = figure()
        used_colors = {}
    else:
        used_colors = dict([(p.get_label(),p.get_facecolor()) for p in fig.axes[0].patches])
    # Evaluate species group name
    # and return a species object
    species_group = mech(species_group)

    # Evaluate process group name
    # and return a process object
    process_obj = mech(process)
    
    # Get the list of component species
    component_species = species_group.names()
        
    # Bar chart starts at the bottom
    bottom, top = 0, 0
    
    # Create a receptacle for values
    vals = []
    
    # For each component species, get the process value
    # Use zero where missing
    for species in component_species:
        factor = species_group[species]
        species_obj = mech(species)
        try:
            vals.append((species, float(factor) * process_obj[species_obj].array().sum()))
        except KeyError, (m):
            warn(m)
            if missing is not None:
                vals.append((species, missing))

    # The number of patches is based on the number of components
    nbars = len(vals)
    colors = iter(get_cmap(cmap)(arange(nbars, dtype = 'f')/(nbars-1)))

    # Normalize to 100% if requested
    if normalize:
        total = process_obj[species_group].array().sum()
        vals = [ (k, v/total) for k, v in vals]

    # Sort based on value if requested
    if sort:
        vals = [ (v, k) for k, v in vals]
        vals.sort(reverse = True)
        vals = [ (k, v) for v, k in vals]

    # Plot each component value using pre-existing
    # color where possible
    for name, val in vals:
        if used_colors.has_key(name):
            color = used_colors[name]
        else:
            while True:
                color = rgb2hex(colors.next()[:-1])
                if color not in used_colors.values():
                    break
            used_colors[name] = color    
        
        bar(loc, val, width, bottom = bottom, label = name, color = color )
        bottom += val
    return fig
    
if __name__ == '__main__':
    from pyPA.utils.util import AttrDict
    from net_balance import get_pure_mech
    from permm.netcdf import NetCDFFile
    conf = AttrDict()
    conf['mechanism'] = 'cb05'
    conf['model'] = 'camx'
    conf['mrgfile'] = '/Users/barronh/Development/net_balance/src/net_balance/testdata/test.mrg.nc'
    conf['species']  = dict(HNO3=dict(ymin=None),
                        NO2=dict(ymin=None),
                        NO=dict(ymin=None),
                        NOx=dict(ymin=None),
                        NOyN=dict(ymin=None),
                        NTR=dict(ymin=None)
                        )
    conf['process'] = {'H_Trans': dict(color = 'b'), 
                    'V_Trans': dict(color = 'g'),
                    'Emissions': dict(color = 'r'),
                    'Deposit': dict(color = 'c'),
                    'Aero_Chem': dict(color = 'm'),
                    'CHEM': dict(color = 'y'),
                    'Motion': dict(color = 'teal'),
                    'TEMPADJ': dict(color = 'aqua')
                    }
    conf['init'] = 'INIT'
    conf['final'] = 'FCONC'
    mech = get_pure_mech('_'.join([conf['mechanism'],conf['model']]))
    mrg_file = NetCDFFile(conf['mrgfile'],'r')
    mech.set_mrg(mrg_file)

    conf['chem'] = 'CHEM'
    conf['title'] = '%(species)s IPR plot'
    conf['outdir'] = '.'
    conf['end_date'] = True
    fig = profile_process(mech, 'NOyN', 'INIT', loc = 0, width = .2, sort = False, missing = None)
    fig = profile_process(mech, 'NOyN', 'Emissions', loc = 0.2, width = .2, sort = False, fig = fig, missing = None)
    fig = profile_process(mech, 'NOyN', 'H_Trans', loc = 0.4, width = .2, sort = False, fig = fig, missing = None)
    fig = profile_process(mech, 'NOyN', 'V_Trans', loc = 0.6, width = .2, sort = False, fig = fig, missing = None)
    fig = profile_process(mech, 'NOyN', 'FCONC', loc = .8, width = .2, sort = False, fig = fig, missing = None)
    patches={}
    for patch in fig.axes[0].patches:
        patches.setdefault(patch.get_label(), patch)
    names = patches.keys()
    patches = [patches[k] for k in names]
    
    fig.legend(patches, names)
    xticks(arange(5, dtype = 'f')/5+.1, ['Initial', 'Emis', 'Horz', 'Vert', 'Final'])
    axis('tight')
    fig.savefig('NOyN_bar.pdf', fmt = 'pdf')


from distutils.core import setup
try:
    from setuptools import setup
except:
    pass
import os
import sys
from warnings import warn

def find_packages():
    import os
    packages = []
    walker = os.walk('src')
    prefix = os.path.join(os.path.curdir,'src')
    for thisdir, itsdirs, itsfiles in walker:
        if '__init__.py' in itsfiles:
            packages.append(thisdir[len(prefix)-1:])
    
    return packages
            
def find_data():
    import os
    import re
    data_pattern = re.compile(r'.*(.|_)(yaml|nc|net|irr|phy|ptb|sum|voc|txt|xls|graffle)$')
    data = []
    prefix = os.path.join(os.path.curdir,'src', 'permm')
    walker = os.walk('src')
    for thisdir, itsdirs, itsfiles in walker:
        if thisdir != os.path.join('src','permm.egg-info'):
            data.extend([os.path.join(thisdir[len(prefix)-1:],f) for f in itsfiles if data_pattern.match(f) is not None])
    
    return data

packages = find_packages()
data = find_data()

setup(name = 'permm',
      version = '1.0',
      author = 'Barron Henderson',
      author_email = 'barronh@gmail.com',
      maintainer = 'Barron Henderson',
      maintainer_email = 'barronh@gmail.com',
      packages = packages,
      package_dir = {'': 'src'},
      package_data = {'permm': data},
      scripts = ['scripts/permm'],
      requires = ['numpy (>=1.2)', 'yaml', 'netCDF4'],
      url = 'http://github.com/barronh/permm/',
      download_url = 'https://github.com/barronh/permm/archive/v1.0.zip'
      )



__all__ = ['Mechanism', 'Species', 'Reaction', 'atoms', 'mechanism_dict']

if __name__ != '__main__':
    from core.Species import Species
    from core.Reaction import Reaction
    from core.Mechanism import Mechanism
    from mechanisms import mechanism_dict, atoms
    import mechanisms
    import getmech
    import Shell
    import GUI
    from getmech import get_pure_mech, get_prepared_mech
    get_mech = get_pure_mech
else:
    from permm.main import parse_and_run
    parse_and_run()


__all__ = ['IRRTable']

def IRRTable(mech, delim = '_'):
    reaction_keys = [(int(k.split(delim)[1]),k) for k in mech.reaction_dict.keys()]
    reaction_keys.sort()
    reaction_keys = [k[1] for k in reaction_keys]
    result = ''
    template = '%s\t%13.7f\t%s\n'
    for rkey in reaction_keys:
        result += template % (rkey, mech.irr[rkey].sum(), mech.nreaction_dict[rkey].sum())
        
    return result